import asyncio
import timeit

import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]

from ahbicht.expressions.ahb_expression_parser import _cache as ahb_expr_cache
from ahbicht.expressions.ahb_expression_parser import _parser as ahb_expr_parser
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import _cache as cond_expr_cache
from ahbicht.expressions.condition_expression_parser import _parser as cond_expr_parser
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree


class TestCaching:
    """
    Tests the caching capabilities of both ahb and condition expression parsers
    """

    @pytest.fixture()
    def clear_caches(self):
        ahb_expr_cache.clear()
        cond_expr_cache.clear()

    # test expressions have to be unique per test case because the tests interfere (they share the same "global" cache)

    @pytest.mark.parametrize(
        "ahb_expression",
        [
            pytest.param("Muss [7] U [8]", id="simple expression"),
            pytest.param(
                "Muss [7] U [8] X ([112] O [2343] X [3] U [9]) O ([1] U [2] U ([12] O [13]) O [12 U [222])",
                id="longer expression",
            ),
        ],
    )
    def test_ahb_expression_cache_performance(self, clear_caches, ahb_expression: str):
        # be careful with performance tests. they naturally behave different on different systems
        calls = 1000
        time_with_cache = timeit.timeit(
            lambda: parse_ahb_expression_to_single_requirement_indicator_expressions(
                ahb_expression, disable_cache=False
            ),
            number=calls,
        )
        time_without_cache = timeit.timeit(
            lambda: parse_ahb_expression_to_single_requirement_indicator_expressions(
                ahb_expression, disable_cache=True
            ),
            number=calls,
        )

        avg_time_per_parsing_with_cache = time_with_cache / calls
        avg_time_per_parsing_without_cache = time_without_cache / calls
        assert avg_time_per_parsing_with_cache < avg_time_per_parsing_without_cache / 100
        # a 100-fold performance improvement for 1000 calls of a simple expression seems fair
        # still, the overall time spent in parsing is not large: the order of magnitude for parsing without cache is 1ms

    def test_ahb_expression_cache_disabled(self, clear_caches, mocker):
        ahb_expression = "Muss [1] U [2]"
        parse_spy = mocker.spy(ahb_expr_parser, "parse")
        for _ in range(100):
            _ = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression, disable_cache=True)
        assert parse_spy.call_count == 100

    def test_ahb_expression_cache_sync(self, clear_caches, mocker):
        ahb_expression = "Muss [3] U [4]"
        parse_spy = mocker.spy(ahb_expr_parser, "parse")
        for _ in range(100):
            _ = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        parse_spy.assert_called_once_with(ahb_expression)

    async def test_ahb_expression_cache_async(self, clear_caches, mocker):

        ahb_expression = "Muss [5] U [6]"
        parse_spy = mocker.spy(ahb_expr_parser, "parse")

        async def parsing_task():
            parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)

        tasks = [parsing_task() for _ in range(100)]
        await asyncio.gather(*tasks)
        parse_spy.assert_called_once_with(ahb_expression)

    def test_condition_expression_cache_sync(self, clear_caches, mocker):
        cond_expression = "[1] U [2]"
        parse_spy = mocker.spy(cond_expr_parser, "parse")
        for _ in range(100):
            _ = parse_condition_expression_to_tree(cond_expression)
        parse_spy.assert_called_once_with(cond_expression)

    async def test_condition_expression_cache_async(self, clear_caches, mocker):
        # test expression has to be different from the cond_expression in the sync test case because the tests interfere
        cond_expression = "[3] U [4]"
        parse_spy = mocker.spy(cond_expr_parser, "parse")

        async def parsing_task():
            parse_condition_expression_to_tree(cond_expression)

        tasks = [parsing_task() for _ in range(100)]
        await asyncio.gather(*tasks)
        parse_spy.assert_called_once_with(cond_expression)



"""
Schemata for the JSON serialization of expressions.
"""

from lark import Tree
from marshmallow import post_dump, pre_load

from ahbicht.json_serialization.tree_schema import TreeSchema


class ConciseTreeSchema(TreeSchema):
    """
    A tree schema that leads to significantly more compact output but can only be used for serialization
    (not deserialization)
    """

    # pylint:disable=unused-argument
    @post_dump
    def serialize(self, data, **kwargs) -> dict:
        """
        Serialize as compact/concise tree
        """
        return _compress(data)

    @pre_load
    def deserialize(self, data, **kwargs) -> Tree:
        """
        This is out of scope
        """
        raise NotImplementedError("Deserializing a compact tree is not supported.")


def _compress(data: dict) -> dict:
    """
    a function that "throws away" unnecessary data.
    The price we pay is that we loose the ability to easily deserialize the result.
    But if we're only interested in a simple tree that's fine.
    """
    if (
        "children" in data
        and "type" in data
        and (data["type"].endswith("_composition") or data["type"].endswith("_expression"))
    ):
        return {data["type"]: [_compress(child) for child in data["children"]]}
    if "tree" in data and "token" in data and data["token"] is None:
        return _compress(data["tree"])
    if "tree" in data and "token" in data and data["tree"] is None:
        return _compress(data["token"])
    if "type" in data and "children" in data:  # and data["type"] in {"MODAL_MARK", "condition_key"}:
        return _compress(data["children"][0]["token"]["value"])
    if "type" in data and data["type"] in {"MODAL_MARK"}:
        return data["value"]
    return data



"""
Test for the expansion of packages.
"""
from typing import Mapping, Optional

import inject
import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]
from maus.edifact import EdifactFormat, EdifactFormatVersion

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableDataProvider
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.expression_resolver import expand_packages
from ahbicht.expressions.package_expansion import JsonFilePackageResolver, PackageResolver
from ahbicht.mapping_results import PackageKeyConditionExpressionMapping
from unittests.defaults import DefaultPackageResolver, return_empty_dummy_evaluatable_data


class TestPackageResolver:
    """
    Test for the expansions of packages
    """

    @pytest.fixture
    def inject_package_resolver(self, request: SubRequest):

        result_dict: Mapping[str, Optional[str]] = request.param

        resolver = DefaultPackageResolver(result_dict)
        inject.clear_and_configure(
            lambda binder: binder.bind(  # type:ignore[arg-type]
                TokenLogicProvider, SingletonTokenLogicProvider([resolver])
            ).bind_to_provider(EvaluatableDataProvider, return_empty_dummy_evaluatable_data)
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        "inject_package_resolver",
        [{"123P": "[1] U ([2] O [3])"}],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "unexpanded_expression, expected_expanded_expression",
        [
            pytest.param("[123P]", "[1] U ([2] O [3])"),
            pytest.param("[123P7..8]", "[1] U ([2] O [3])"),
            pytest.param("[123P10..11]", "[1] U ([2] O [3])"),
            pytest.param("[123P0..5]", "[1] U ([2] O [3])"),
            pytest.param("[17] U [123P]", "[17] U ([1] U ([2] O [3]))"),
        ],
    )
    async def test_correct_injection(
        self, inject_package_resolver, unexpanded_expression: str, expected_expanded_expression: str
    ):
        unexpanded_tree = parse_condition_expression_to_tree(unexpanded_expression)
        actual_tree = await expand_packages(parsed_tree=unexpanded_tree)
        assert actual_tree is not None
        expected_expanded_tree = parse_condition_expression_to_tree(expected_expanded_expression)
        assert actual_tree == expected_expanded_tree

    @pytest.mark.parametrize(
        "inject_package_resolver",
        [{"123P": "[1] U ([2] O [3])"}],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "unexpanded_expression, error_message",
        [
            pytest.param("[123P8..7]", "0≤n≤m is not fulfilled for n=8, m=7"),
        ],
    )
    async def test_invalid_package_repeatability(
        self, inject_package_resolver, unexpanded_expression: str, error_message: str
    ):
        unexpanded_tree = parse_condition_expression_to_tree(unexpanded_expression)
        with pytest.raises(ValueError) as invalid_repeatability_error:
            _ = await expand_packages(parsed_tree=unexpanded_tree)
        assert error_message in str(invalid_repeatability_error)

    @pytest.mark.datafiles("./unittests/provider_test_files/example_package_mapping_dict.json")
    @pytest.mark.datafiles("./unittests/provider_test_files/example_package_mapping_list.json")
    @pytest.mark.parametrize(
        "filename",
        [pytest.param("example_package_mapping_dict.json"), pytest.param("example_package_mapping_list.json")],
    )
    async def test_file_based_package_resolver(self, filename, datafiles):
        """Tests if package resolver provider is instantiated correctly."""
        path_to_hint_json = datafiles / filename
        package_resolver: PackageResolver = JsonFilePackageResolver(
            edifact_format=EdifactFormat.UTILMD,
            edifact_format_version=EdifactFormatVersion.FV2104,
            file_path=path_to_hint_json,
        )
        assert package_resolver.edifact_format == EdifactFormat.UTILMD
        assert package_resolver.edifact_format_version == EdifactFormatVersion.FV2104
        assert await package_resolver.get_condition_expression("123P") == PackageKeyConditionExpressionMapping(
            edifact_format=EdifactFormat.UTILMD, package_key="123P", package_expression="[1] U [2] O [3]"
        )
        assert await package_resolver.get_condition_expression("456P") == PackageKeyConditionExpressionMapping(
            edifact_format=EdifactFormat.UTILMD, package_key="456P", package_expression="[4] U [5] X [6]"
        )



"""
contains reusable classes and instances of AHBicht Evaluators and Resolvers.
They are supposed to be used in unittests where they need to be injected but don't have a real purpose.
Inject them to have a concise test setup.
"""
from maus.edifact import EdifactFormat, EdifactFormatVersion

#: the default edifact format used in the unit tests
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.expressions.hints_provider import DictBasedHintsProvider
from ahbicht.expressions.package_expansion import DictBasedPackageResolver

default_test_format: EdifactFormat = EdifactFormat.UTILMD
#: the default edifact format version used in the unit tests
default_test_version: EdifactFormatVersion = EdifactFormatVersion.FV2210
#: an empty EvaluatableData instance
empty_default_test_data: EvaluatableData = EvaluatableData(
    edifact_seed={}, edifact_format=default_test_format, edifact_format_version=default_test_version
)


class EmptyDefaultRcEvaluator(RcEvaluator):
    """
    An RC Evaluator in the default edifact format and edifact format version
    """

    def _get_default_context(self):
        return None

    def __init__(self):
        super().__init__()
        self.edifact_format = default_test_format
        self.edifact_format_version = default_test_version


empty_default_rc_evaluator = EmptyDefaultRcEvaluator()


class EmptyDefaultFcEvaluator(FcEvaluator):
    """
    An (empty) FC Evaluator in the default edifact format and edifact format version
    """

    def __init__(self):
        super().__init__()
        self.edifact_format = default_test_format
        self.edifact_format_version = default_test_version


empty_default_fc_evaluator = EmptyDefaultFcEvaluator()


class DefaultHintsProvider(DictBasedHintsProvider):
    """
    An (empty) Hints Provider in the default edifact format and edifact format version
    """

    def __init__(self, mappings):
        super().__init__(mappings)
        self.edifact_format = default_test_format
        self.edifact_format_version = default_test_version


empty_default_hints_provider = DefaultHintsProvider({})


class DefaultPackageResolver(DictBasedPackageResolver):
    """
    An (empty) Package Resolver in the default edifact format and edifact format version
    """

    def __init__(self, mappings):
        super().__init__(mappings)
        self.edifact_format = default_test_format
        self.edifact_format_version = default_test_version


empty_default_package_resolver = DefaultPackageResolver({})


def return_empty_dummy_evaluatable_data() -> EvaluatableData:
    """
    :return: empty evaluatable data in the default format and format version
    """
    return empty_default_test_data



"""
Tests the evaluation of the start/end of a German Strom- or Gastag.
"""
from datetime import datetime, timedelta, timezone

import pytest  # type:ignore[import]

from ahbicht.content_evaluation.german_strom_and_gas_tag import (
    berlin,
    is_gastag_limit,
    is_stromtag_limit,
    parse_as_datetime,
)


class TestGermanStromAndGasTag:
    @pytest.mark.parametrize(
        "dt_string, expected_datetime",
        [
            pytest.param(
                "2019-12-31T23:00:00+01:00", datetime(2019, 12, 31, 23, 0, 0, tzinfo=timezone(timedelta(seconds=3600)))
            ),
            pytest.param("2022-01-01T00:00:00+00:00", datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc)),
            pytest.param("2022-01-01T00:00:00Z", datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc)),
            pytest.param("2021-12-31T23:00:00-01:00", datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc)),
        ],
    )
    def test_successful_parsing(self, dt_string: str, expected_datetime: datetime):
        actual, error = parse_as_datetime(dt_string)
        assert error is None
        assert actual == expected_datetime

    @pytest.mark.parametrize(
        "dt_string, expected_error_msg",
        [
            pytest.param("2019-12-31T00:00:00", "Neither offset nor timezone was given"),
            pytest.param("2019-12-31T25:00:00+00:00", "hour must be in 0..23"),
            pytest.param("foo", "Invalid isoformat string"),
            pytest.param("", "empty or None"),
        ],
    )
    def test_errornous_parsing(self, dt_string: str, expected_error_msg: str):
        actual, error = parse_as_datetime(dt_string)
        assert actual is None
        assert error is not None
        assert error.format_constraint_fulfilled is False
        assert expected_error_msg in error.error_message  # type:ignore[operator]

    @pytest.mark.parametrize(
        "dt, expected_is_start_or_end_of_german_stromtag",
        [
            pytest.param(datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc), False),
            pytest.param(datetime(2019, 12, 31, 23, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime(2019, 12, 31, 22, 0, 0, tzinfo=timezone.utc), False),
            pytest.param(datetime(2010, 1, 1, 0, 0, 0, tzinfo=berlin), True),
            pytest.param(datetime(2010, 1, 1, 0, 0, 0, tzinfo=timezone(timedelta(seconds=3600))), True),
            pytest.param(datetime(2022, 3, 26, 23, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-03-27T00:00:00+01:00"), True),
            pytest.param(datetime.fromisoformat("2022-03-27T00:00:00+02:00"), False),
            pytest.param(datetime(2022, 3, 27, 22, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-03-28T00:00:00+02:00"), True),
            pytest.param(datetime(2022, 10, 29, 22, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-10-30T00:00:00+02:00"), True),
            pytest.param(datetime(2022, 10, 30, 23, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-10-31T00:00:00+01:00"), True),
            pytest.param(datetime.fromisoformat("2022-10-31T00:00:00+02:00"), False),
            pytest.param(datetime.fromisoformat("2022-10-29T12:00:00-10:00"), True, id="Hawaii, German DST"),
            pytest.param(datetime.fromisoformat("2022-10-30T13:00:00-10:00"), True, id="Hawaii, German standard time"),
            pytest.param(datetime.fromisoformat("2022-10-30T07:00:00+09:00"), True, id="Tokyo, German DST"),
            pytest.param(datetime.fromisoformat("2022-10-31T08:00:00+09:00"), True, id="Tokyo, German standard time"),
        ],
    )
    def test_stromtag(self, dt: datetime, expected_is_start_or_end_of_german_stromtag: bool):
        actual = is_stromtag_limit(dt)
        assert actual == expected_is_start_or_end_of_german_stromtag

    @pytest.mark.parametrize(
        "dt, expected_is_start_or_end_of_german_gastag",
        [
            pytest.param(datetime(2020, 1, 1, 0, 0, 0, tzinfo=timezone.utc), False),
            pytest.param(datetime(2020, 1, 1, 5, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime(2020, 1, 1, 4, 0, 0, tzinfo=timezone.utc), False),
            pytest.param(datetime(2022, 3, 26, 5, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-03-26T06:00:00+01:00"), True),
            pytest.param(datetime.fromisoformat("2022-03-27T04:00:00+02:00"), False),
            pytest.param(datetime(2022, 3, 27, 4, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-03-27T06:00:00+02:00"), True),
            pytest.param(datetime(2022, 10, 29, 4, 0, 0, 0, tzinfo=timezone.utc), True),
            pytest.param(datetime.fromisoformat("2022-10-30T06:00:00+01:00"), True),
            pytest.param(datetime.fromisoformat("2022-10-30T06:00:00+02:00"), False),
            pytest.param(datetime.fromisoformat("2022-10-29T19:00:00-10:00"), True, id="Hawaii, German DST"),
            pytest.param(datetime.fromisoformat("2022-10-30T19:00:00-10:00"), True, id="Hawaii, German standard time"),
            pytest.param(datetime.fromisoformat("2022-10-29T09:45:00+05:45"), True, id="Nepal, German DST"),
            pytest.param(datetime.fromisoformat("2022-10-30T10:45:00+05:45"), True, id="Nepal, German standard time 1"),
            pytest.param(datetime.fromisoformat("2022-10-31T10:45:00+05:45"), True, id="Nepal, German standard time 2"),
        ],
    )
    def test_gastag(self, dt: datetime, expected_is_start_or_end_of_german_gastag: bool):
        actual = is_gastag_limit(dt)
        assert actual == expected_is_start_or_end_of_german_gastag



"""
This module contains the classes for the evaluation results.
A "result" is the outcome of an evaluation. It requires actual data to be present.
"""

# pylint: disable=too-few-public-methods, no-member,  unused-argument
from typing import Optional

import attrs
from marshmallow import Schema, fields, post_load

from ahbicht.expressions.enums import RequirementIndicator, RequirementIndicatorSchema


@attrs.define(auto_attribs=True, kw_only=True)
class RequirementConstraintEvaluationResult:
    """
    A class for the result of the requirement constraint evaluation.
    """

    #: true if condition expression in regard to requirement constraints evaluates to true
    requirement_constraints_fulfilled: bool = attrs.field(validator=attrs.validators.instance_of(bool))
    #: true if it is dependent on requirement constraints
    requirement_is_conditional: bool = attrs.field(validator=attrs.validators.instance_of(bool))

    format_constraints_expression: Optional[str] = attrs.field(
        default=None, validator=attrs.validators.optional(attrs.validators.instance_of(str))
    )
    #: Hint text that should be displayed in the frontend, e.g. "[501] Hinweis: 'ID der Messlokation'"
    hints: Optional[str] = attrs.field(
        default=None, validator=attrs.validators.optional(attrs.validators.instance_of(str))
    )


class RequirementConstraintEvaluationResultSchema(Schema):
    """
    A schema to (de-)serialize RequirementConstraintEvaluationResult
    """

    requirement_constraints_fulfilled = fields.Boolean()
    requirement_is_conditional = fields.Boolean()

    format_constraints_expression = fields.String(load_default=None)
    hints = fields.String(load_default=None)

    @post_load
    def deserialize(self, data, **kwargs) -> RequirementConstraintEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual RequirementConstraintEvaluationResult
        :param data:
        :param kwargs:
        :return:
        """
        return RequirementConstraintEvaluationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class FormatConstraintEvaluationResult:
    """
    A class for the result of the format constraint evaluation.
    """

    #: true if data entered obey the format constraint expression
    format_constraints_fulfilled: bool = attrs.field(validator=attrs.validators.instance_of(bool))

    #: All error messages that lead to not fulfilling the format constraint expression
    error_message: Optional[str] = attrs.field(
        default=None, validator=attrs.validators.optional(attrs.validators.instance_of(str))
    )


class FormatConstraintEvaluationResultSchema(Schema):
    """
    A class to (de-)serialize FormatConstraintEvaluationResult
    """

    format_constraints_fulfilled = fields.Boolean()
    error_message = fields.String(allow_none=True, load_default=None)

    @post_load
    def deserialize(self, data, **kwargs) -> FormatConstraintEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual FormatConstraintEvaluationResult
        :param data:
        :param kwargs:
        :return:
        """
        return FormatConstraintEvaluationResult(**data)


@attrs.define(auto_attribs=True, kw_only=True)
class AhbExpressionEvaluationResult:
    """
    A class for the result of an ahb expression evaluation.
    """

    requirement_indicator: RequirementIndicator  # i.e. "Muss", "M", "Soll", "S", "Kann", "K", "X", "O", "U"
    requirement_constraint_evaluation_result: RequirementConstraintEvaluationResult
    format_constraint_evaluation_result: FormatConstraintEvaluationResult


class AhbExpressionEvaluationResultSchema(Schema):
    """
    A schema to (de-)serialize AhbExpressionEvaluationResults
    """

    requirement_indicator = fields.Nested(RequirementIndicatorSchema)  # because union is not supported by marshmallow
    requirement_constraint_evaluation_result = fields.Nested(RequirementConstraintEvaluationResultSchema())
    format_constraint_evaluation_result = fields.Nested(FormatConstraintEvaluationResultSchema())

    @post_load
    def deserialize(self, data, **kwargs) -> AhbExpressionEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual AhbExpressionEvaluationResult
        """
        return AhbExpressionEvaluationResult(**data)



"""
Functions that are not clearly related to another module
"""
import asyncio
import inspect
from typing import Awaitable, List, TypeVar, Union

Result = TypeVar("Result")


async def gather_if_necessary(results_and_awaitable_results: List[Union[Result, Awaitable[Result]]]) -> List[Result]:
    """
    Await the awaitables, pass the un-awaitable results
    :param results_and_awaitable_results: heterogenous list of both Ts and Awaitable[T]s.
    :return: list of T in the same order as in the input param.
    """
    awaitable_indexes = [n for n, x in enumerate(results_and_awaitable_results) if inspect.isawaitable(x)]
    awaited_results = await asyncio.gather(*[x for x in results_and_awaitable_results if inspect.isawaitable(x)])
    result: List[Result] = []
    awaited_results_index = 0
    for index, obj in enumerate(results_and_awaitable_results):
        if index in awaitable_indexes:
            result.append(awaited_results[awaited_results_index])
            awaited_results_index += 1
        else:
            # we are sure obj is of type T
            result.append(obj)  # type:ignore[arg-type]
    return result



import asyncio
from typing import Dict, Optional

import pytest  # type:ignore[import]

from ahbicht.content_evaluation import fc_evaluators
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint


class _MyFcEvaluator(FcEvaluator):
    def _evaluate_has_length(self, entered_input: Optional[str], length: int) -> EvaluatedFormatConstraint:
        """
        check if input is length character long
        """
        if not entered_input:
            return EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="Input is None or Empty")
        if len(entered_input) == length:
            return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)
        return EvaluatedFormatConstraint(
            format_constraint_fulfilled=False,
            error_message=f"Input has length {len(entered_input)} but expected {length}",
        )

    def evaluate_1(self, entered_input: Optional[str]):
        """
        check if input is 1 character long
        """
        return self._evaluate_has_length(entered_input, 1)

    def evaluate_2(self, entered_input: Optional[str]):
        """
        check if input is 2 characters long
        """
        return self._evaluate_has_length(entered_input, 2)

    async def evaluate_3(self, entered_input: Optional[str]):
        """
        check if input is 3 character long
        """
        return self._evaluate_has_length(entered_input, 3)

    async def evaluate_4(self, entered_input: Optional[str]):
        """
        check if input is 4 character long
        """
        return self._evaluate_has_length(entered_input, 4)


def _build_expectations(actual_length: int) -> Dict[str, EvaluatedFormatConstraint]:
    result = {}
    for length in range(1, 5):
        error_message: Optional[str]
        if actual_length == 0:
            error_message = "Input is None or Empty"
        elif actual_length == length:
            error_message = None
        else:
            error_message = f"Input has length {actual_length} but expected {length}"

        value = EvaluatedFormatConstraint(
            format_constraint_fulfilled=length == actual_length, error_message=error_message
        )
        result.update({str(length): value})
    return result


class TestFormatConstraintsContextVar:
    """Tests that the context variable used by the FC Evaluators works as designed"""

    async def test_context_var_is_context_sensitive(self):
        evaluator = _MyFcEvaluator()
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set("something to confuse the evaluation?")

        async def first_evaluation():
            fc_evaluators.text_to_be_evaluated_by_format_constraint.set("a")
            assert await evaluator.evaluate_format_constraints(["1", "2", "3", "4"]) == _build_expectations(1)

        async def second_evaluation():
            fc_evaluators.text_to_be_evaluated_by_format_constraint.set("bb")
            assert await evaluator.evaluate_format_constraints(["1", "2", "3", "4"]) == _build_expectations(2)

        async def third_evaluation():
            fc_evaluators.text_to_be_evaluated_by_format_constraint.set("ccc")
            assert await evaluator.evaluate_format_constraints(["1", "2", "3", "4"]) == _build_expectations(3)

        async def forth_evaluation():
            fc_evaluators.text_to_be_evaluated_by_format_constraint.set("dddd")
            assert await evaluator.evaluate_format_constraints(["1", "2", "3", "4"]) == _build_expectations(4)

        await asyncio.gather(*[first_evaluation(), second_evaluation(), third_evaluation(), forth_evaluation()])



""" Test for the evaluation of the format constraint expression. """
from typing import Optional

import inject
import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]

from ahbicht.content_evaluation import fc_evaluators
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableDataProvider
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.evaluation_results import FormatConstraintEvaluationResult
from ahbicht.expressions.condition_nodes import EvaluatedFormatConstraint
from ahbicht.expressions.format_constraint_expression_evaluation import (
    _build_evaluated_format_constraint_nodes,
    format_constraint_evaluation,
)
from unittests.defaults import default_test_format, default_test_version, return_empty_dummy_evaluatable_data


class DummyFcEvaluator(FcEvaluator):
    """
    A dummy Format Constraint Evaluator
    """

    edifact_format = default_test_format
    edifact_format_version = default_test_version

    async def evaluate_950(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        [950] Format: Marktlokations-ID
        """
        # this is just a minimal working example; we skip all the stuff like check digits and so on for simplicity
        is_malo: bool = entered_input and len(entered_input) == 11  # type:ignore[assignment]
        if is_malo:
            error_message = None
        else:
            error_message = f"'{entered_input}' is not 11 characters long and hence no MaLo."
        return EvaluatedFormatConstraint(format_constraint_fulfilled=is_malo, error_message=error_message)

    async def evaluate_951(self, entered_input: str) -> EvaluatedFormatConstraint:
        """
        [951] Format: Zählpunktbezeichnung
        """
        # this is just a minimal working example; we skip regex matching and integrity checks for simplicity
        is_zaehlpunkt: bool = entered_input and len(entered_input) == 33  # type:ignore[assignment]
        if is_zaehlpunkt:
            error_message = None
        else:
            error_message = f"'{entered_input}' is not a valid Zählpunktbezeichnung."
        return EvaluatedFormatConstraint(format_constraint_fulfilled=is_zaehlpunkt, error_message=error_message)


class TestFormatConstraintExpressionEvaluation:
    """Test for the evaluation of the format constraint expressions"""

    _input_values = {
        "901": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="902 muss erfüllt sein"),
        "903": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
        "904": EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="904 muss erfüllt sein"),
    }

    @pytest_asyncio.fixture()
    def setup_and_teardown_injector(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(
                TokenLogicProvider, SingletonTokenLogicProvider([DummyFcEvaluator()])
            ).bind_to_provider(EvaluatableDataProvider, return_empty_dummy_evaluatable_data)
        )
        yield
        inject.clear()

    @pytest.mark.parametrize(
        "format_constraint_expression, expected_format_constraints_fulfilled, expected_error_message",
        [
            pytest.param("[901]", True, None),
            pytest.param("[902]", False, "902 muss erfüllt sein"),
            pytest.param("[901]U[903]", True, None),
            pytest.param("[901]U[902]", False, "902 muss erfüllt sein"),
            pytest.param("[902]U[901]", False, "902 muss erfüllt sein"),
            pytest.param("[902]U[904]", False, "'902 muss erfüllt sein' und '904 muss erfüllt sein'"),
            pytest.param("[901]O[903]", True, None),
            pytest.param("[901]O[902]", True, None),
            pytest.param("[902]O[901]", True, None),
            pytest.param("[902]O[904]", False, "'902 muss erfüllt sein' oder '904 muss erfüllt sein'"),
            pytest.param(
                "[901]X[903]", False, "Zwei exklusive Formatdefinitionen dürfen nicht gleichzeitig erfüllt sein"
            ),
            pytest.param("[901]X[902]", True, None),
            pytest.param("[902]X[901]", True, None),
            pytest.param("[902]X[904]", False, "Entweder '902 muss erfüllt sein' oder '904 muss erfüllt sein'"),
            # Tests 'and before or'
            pytest.param("[902]U[904]O[901]", True, None),
            pytest.param("[901]O[902]U[904]", True, None),
            pytest.param("[902]U[901]O[903]", True, None),
            pytest.param("[902]O[901]U[903]", True, None),
            pytest.param("[902]O[901]U[904]", False, "'902 muss erfüllt sein' oder '904 muss erfüllt sein'"),
            pytest.param("[901]U[902]U[903]U[901]", False, "902 muss erfüllt sein"),
            pytest.param("[901]U[903]U[902]U[901]", False, "902 muss erfüllt sein"),
            # a very long one
            pytest.param("[901]U[902]O[901]U[901]U[902]O[902]O[901]", True, None),
            # with brackets
            pytest.param("([902]U[904])O[901]", True, None),
            pytest.param("[902]U([904]O[901])", False, "902 muss erfüllt sein"),
        ],
    )
    async def test_evaluate_valid_format_constraint_expression(
        self, mocker, format_constraint_expression, expected_format_constraints_fulfilled, expected_error_message
    ):
        """
        Tests that valid format_constraint expressions are evaluated as expected.
        Odd condition_keys are True, even condition_keys are False
        """
        mocker.patch(
            "ahbicht.expressions.format_constraint_expression_evaluation._build_evaluated_format_constraint_nodes",
            return_value=self._input_values,
        )

        result = await format_constraint_evaluation(format_constraint_expression)

        assert isinstance(result, FormatConstraintEvaluationResult)
        assert result.format_constraints_fulfilled == expected_format_constraints_fulfilled
        assert result.error_message == expected_error_message

    @pytest.mark.parametrize(
        "format_constraints_expression, input_values, expected_error_message",
        [
            pytest.param(
                "[1]",
                {"1": "no_evaluated_format_constraint"},
                "Please make sure that the passed values are EvaluatedFormatConstraints.",
            ),
            pytest.param(
                "[1]U[2]",
                {"1": EvaluatedFormatConstraint(format_constraint_fulfilled=True)},
                "Please make sure that the input values contain all necessary condition_keys.",
            ),
        ],
    )
    async def test_evaluate_format_constraint_expressions_with_invalid_values(
        self,
        mocker,
        format_constraints_expression: str,
        input_values: dict,
        expected_error_message: str,
    ):
        """Tests that an error is raised when trying to pass invalid values."""
        mocker.patch(
            "ahbicht.expressions.format_constraint_expression_evaluation._build_evaluated_format_constraint_nodes",
            return_value=input_values,
        )

        with pytest.raises(ValueError) as excinfo:
            await format_constraint_evaluation(format_constraints_expression)

        assert expected_error_message in str(excinfo.value)

    @pytest.mark.parametrize(
        "condition_keys, entered_input, expected_evaluated_fc_nodes",
        [
            pytest.param(
                ["950", "951"],
                "12345678913",
                {
                    "950": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
                    "951": EvaluatedFormatConstraint(
                        format_constraint_fulfilled=False,
                        error_message="'12345678913' is not a valid Zählpunktbezeichnung.",
                    ),
                },
            ),
            pytest.param(
                ["950", "951"],
                "DE00056266802AO6G56M11SN51G21M24S",
                {
                    "950": EvaluatedFormatConstraint(
                        format_constraint_fulfilled=False,
                        error_message="'DE00056266802AO6G56M11SN51G21M24S' is not 11 characters long and hence no MaLo.",
                    ),
                    "951": EvaluatedFormatConstraint(format_constraint_fulfilled=True),
                },
            ),
        ],
    )
    async def test_build_evaluated_format_constraint_nodes(
        self, condition_keys, entered_input, expected_evaluated_fc_nodes, setup_and_teardown_injector
    ):
        """Tests that evaluated format constraints nodes are build correctly."""
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set(entered_input)
        evaluated_fc_nodes = await _build_evaluated_format_constraint_nodes(condition_keys)
        assert evaluated_fc_nodes == expected_evaluated_fc_nodes

    @pytest.mark.parametrize(
        "format_constraint_expression, entered_input, is_successful, error_message",
        [
            pytest.param("[931]", "2022-01-01T00:00:00+00:00", True, "+00:00"),
            pytest.param("[931]", "2022-01-01T00:00:00Z", True, "Z is +00:00"),
            pytest.param("[931]", None, False, "None"),
            pytest.param("[931]", "2022-12-31T16:00:00-08:00", False, None),  # yes, it's truly a format constraint
            pytest.param("[931]", "2022-01-01T01:00:00+01:00", False, None),  # yes, it's that bad
            pytest.param("[932]", None, False, "empty or None"),
            pytest.param("[932]", "2022-01-01T00:00:00Z", False, "Stromtag"),
            pytest.param("[933]", "2022-01-01T00:00:00Z", False, "Stromtag"),
            pytest.param("[934]", "2022-01-01T06:00:00Z", False, "Gastag"),
            pytest.param("[935]", "2022-01-01T06:00:00Z", False, "Gastag"),
            pytest.param("[932]", "2022-06-01T00:00:00+02:00", True, None),
            pytest.param("[933]", "2022-01-01T00:00:00+01:00", True, None),
            pytest.param("[934]", "2022-06-01T06:00:00+02:00", True, None),
            pytest.param("[935]", "2022-01-01T06:00:00+01:00", True, None),
        ],
    )
    async def test_93x_format_constraints(
        self,
        format_constraint_expression,
        entered_input: str,
        is_successful: bool,
        error_message: Optional[str],
        setup_and_teardown_injector,
    ):
        """
        Tests that the default FC evaluator ships evaluation methods for 932, 933, 934 and 935 (those expanded from UBx)
        """
        fc_evaluators.text_to_be_evaluated_by_format_constraint.set(entered_input)
        result = await format_constraint_evaluation(format_constraint_expression)
        assert result is not None
        assert result.format_constraints_fulfilled == is_successful
        if is_successful is False and error_message is not None:
            assert error_message in result.error_message  # type:ignore[operator]



"""
A module to allow easy distinction between different types of condition nodes (by mapping their integer key)
"""
from enum import Enum


class ConditionNodeType(str, Enum):
    """
    Possible types of condition nodes.
    The value is usually determined using the key of the respective condition node.
    """

    FORMAT_CONSTRAINT = "FORMAT_CONSTRAINT"  #: a format constraint
    REQUIREMENT_CONSTRAINT = "REQUIREMENT_CONSTRAINT"  #: a requirement constraint
    HINT = "HINT"  #: a hint
    PACKAGE = "PACKAGE"  #: a package
    REPEATABILITY_CONSTRAINT = "REPEATABILITY_CONSTRAINT"  #: a repeatability constraint


def derive_condition_node_type(condition_key: str) -> ConditionNodeType:
    """
    Returns the corresponding type of condition node for a given condition key
    """
    if condition_key.endswith("P"):
        return ConditionNodeType.PACKAGE
        # todo: implement wiederholbarkeiten https://github.com/Hochfrequenz/ahbicht/issues/96
    if 1 <= int(condition_key) <= 499:
        return ConditionNodeType.REQUIREMENT_CONSTRAINT
    if 500 <= int(condition_key) <= 900:
        return ConditionNodeType.HINT
    if 901 <= int(condition_key) <= 999:
        return ConditionNodeType.FORMAT_CONSTRAINT
    if 2000 <= int(condition_key) <= 2499:
        return ConditionNodeType.REPEATABILITY_CONSTRAINT
    raise ValueError("Condition key is not in valid number range.")



"""
Contains the CategorizedKeyExtract and a schema for (de)serialization.
"""
from itertools import combinations, product
from typing import List

import attrs
from marshmallow import Schema, fields, post_load

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint


# pylint: disable=too-few-public-methods,  unused-argument
@attrs.define(auto_attribs=True)
class CategorizedKeyExtract:
    """
    A Categorized Key Extract contains those condition keys that are contained inside an expression.
    For expressions (that do not contain any unresolved package) it's possible to pre-generate all possible outcomes of
    a content evaluation. CategorizedKeyExtract is also the answer to the question:
    'Which information do I need to provide in a ContentEvaluationResult in order to evaluate a given expression?'
    """

    #: list of keys for which you'll need to provide hint texts in a ContentEvaluationResult
    hint_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str), iterable_validator=attrs.validators.instance_of(list)
        )
    )
    #: list of keys for which you'll need to provide EvaluatedFormatConstraints
    format_constraint_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str), iterable_validator=attrs.validators.instance_of(list)
        )
    )

    #: list of keys for which you'll need to provide ConditionFulfilledValues
    requirement_constraint_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(str), iterable_validator=attrs.validators.instance_of(list)
        )
    )

    #: list of packages that need to be resolved (additionally)
    package_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.matches_re(r"^\d+P$"),
            iterable_validator=attrs.validators.instance_of(list)
            # todo: implement wiederholbarkeiten
        )
    )

    #: a list of time conditions, if present
    time_condition_keys: List[str] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.matches_re(r"^UB(?:1|2|3)$"),
            iterable_validator=attrs.validators.instance_of(list),
        )
    )

    def _remove_duplicates(self) -> None:
        """
        remove duplicates from all lists
        """
        self.hint_keys = [*set(self.hint_keys)]
        self.format_constraint_keys = [*set(self.format_constraint_keys)]
        self.requirement_constraint_keys = [*set(self.requirement_constraint_keys)]
        self.package_keys = [*set(self.package_keys)]
        self.time_condition_keys = [*set(self.time_condition_keys)]

    def _sort_keys(self) -> None:
        """
        sort the keys in all lists in ascending order
        """
        self.hint_keys.sort(key=int)
        self.format_constraint_keys.sort(key=int)
        self.requirement_constraint_keys.sort(key=int)
        self.package_keys.sort()
        self.time_condition_keys.sort()

    def sanitize(self) -> None:
        """
        Sanitize the result (remove duplicates, sort keys)
        """
        self._remove_duplicates()
        self._sort_keys()

    def generate_possible_content_evaluation_results(self) -> List[ContentEvaluationResult]:
        """
        A categorized key extract allows generating nearly all possible content evaluation results,
        except for hints, error messages, resolving packages.
        """
        results: List[ContentEvaluationResult] = []
        if len(self.format_constraint_keys) == 0 and len(self.requirement_constraint_keys) == 0:
            return results
        # for easier debugging below, replace the generators "(" with a materialized lists "["
        if len(self.format_constraint_keys) > 0:
            possible_fcs = (
                z
                for z in combinations(
                    product(self.format_constraint_keys, [True, False]), len(self.format_constraint_keys)
                )
                # y[0] is the key (and y[1] is the value (true/false))
                # We only want those combinations where the number of distinct keys == the number of keys present.
                # To better understand this, try evaluating the following expression in your debugger:
                #   list(combinations(product(["A", "B", "C"], [True, False]), len(["A", "B", "C"])))
                # In this (unfiltered) result, you'll find entries with the keys: ('A', 'A', 'B') or ('A', 'B', 'B').
                # These artefacts will be removed with the following if. Only the ('A', 'B', 'C') entries will pass.
                if len({y[0] for y in z}) == len(self.format_constraint_keys)
            )
        else:
            possible_fcs = [(("fc_dummy", True),)]  # type:ignore[assignment]

        if len(self.requirement_constraint_keys) > 0:
            possible_rcs = (
                z
                for z in combinations(
                    product(self.requirement_constraint_keys, ConditionFulfilledValue),
                    len(self.requirement_constraint_keys),
                )
                if len({y[0] for y in z}) == len(self.requirement_constraint_keys)
            )
        else:
            possible_rcs = [(("rc_dummy", ConditionFulfilledValue.NEUTRAL),)]  # type:ignore[assignment]

        for fc_rc_tuple in product(possible_fcs, possible_rcs):
            # This product would have length 0 if one of the "factors" had length 0.
            # In order to prevent 'results' to be empty if either the RC or FC list is empty, we added the 'dummy's.
            result = ContentEvaluationResult(
                hints={hint_key: f"Hinweis {hint_key}" for hint_key in self.hint_keys},
                # kvp is short for key value pair
                format_constraints={
                    fc_kvp[0]: EvaluatedFormatConstraint(format_constraint_fulfilled=fc_kvp[1])
                    for fc_kvp in fc_rc_tuple[0]
                    if fc_kvp[0] != "fc_dummy"
                },
                requirement_constraints={rc_kvp[0]: rc_kvp[1] for rc_kvp in fc_rc_tuple[1] if rc_kvp[0] != "rc_dummy"},
                packages={},  # is always empty because looping over all packages does not make sense in this context
            )
            results.append(result)
        return results


class CategorizedKeyExtractSchema(Schema):
    """
    A schema to (de)serialize CategorizedKeyExtractSchema
    """

    hint_keys = fields.List(fields.String())
    format_constraint_keys = fields.List(fields.String())
    requirement_constraint_keys = fields.List(fields.String())
    package_keys = fields.List(fields.String())
    time_condition_keys = fields.List(fields.String())

    @post_load
    def deserialize(self, data, **kwargs) -> CategorizedKeyExtract:
        """
        Converts the barely typed data dictionary into an actual CategorizedKeyExtractSchema
        """
        return CategorizedKeyExtract(**data)
